@@ -1,7 +1,5 @@
 # -*- coding: utf-8 -*-
 
-# -*- coding: utf-8 -*-
-
 """
 Copyright (c) 2018 Keijack Wu
 
@@ -32,17 +30,16 @@ import threading
 import http.cookies as cookies
 import datetime
 
-from collections import OrderedDict
+from typing import Dict, List, Union
 
-from urllib.parse import unquote
-from typing import Dict, List
-
-from simple_http_server import ModelDict, RegGroup, RegGroups, HttpError, StaticFile, \
+from simple_http_server import ModelDict, Environment, RegGroup, RegGroups, HttpError, StaticFile, \
     Headers, Redirect, Response, Cookies, Cookie, JSONBody, Header, Parameters, PathValue, \
     Parameter, MultipartFile, Request, Session, ControllerFunction, _get_session_factory
 from simple_http_server._http_session_local_impl import SESSION_COOKIE_NAME
+import simple_http_server.__utils as utils
 
 from .logger import get_logger
+from simple_http_server.__utils import get_function_args, get_function_kwargs
 
 _logger = get_logger("http_request_handler")
 
@@ -78,11 +75,11 @@ class ResponseWrapper(Response):
     def is_sent(self) -> bool:
         return self.__is_sent
 
-    def send_error(self, status_code: int, message: str = "") -> None:
-        self.status_code = status_code
-        msg = message if message is not None else ""
-        self.body = {"error": msg}
-        self.send_response()
+    def send_error(self, status_code: int, message: str = "", explain: str = "") -> None:
+        with self.__send_lock__:
+            self.__is_sent = True
+            self.status_code = status_code
+            self.__req_handler.send_error(self.status_code, message=message, explain=explain, headers=self.headers)
 
     def send_redirect(self, url: str) -> None:
         self.status_code = 302
@@ -207,7 +204,7 @@ class FilterContex:
         return status_code, headers, cks, body
 
     def __prepare_args(self):
-        args = _get_args_(self.__controller.func)
+        args = get_function_args(self.__controller.func)
         arg_vals = []
         if len(args) > 0:
             ctr_obj = self.__controller.ctrl_object
@@ -217,7 +214,7 @@ class FilterContex:
         for arg, arg_type_anno in args:
             if arg not in self.request.parameter.keys() \
                     and arg_type_anno not in (Request, Session, Response, RegGroups, RegGroup, Headers, cookies.BaseCookie, cookies.SimpleCookie, Cookies, PathValue, JSONBody, ModelDict):
-                raise HttpError(400, f"Parameter[{arg}] is required! ")
+                raise HttpError(400, "Missing Paramter", f"Parameter[{arg}] is required! ")
             param = self.__get_params_(arg, arg_type_anno)
             arg_vals.append(param)
 
@@ -239,6 +236,8 @@ class FilterContex:
             param = Headers(self.request.headers)
         elif arg_type == RegGroups:
             param = RegGroups(self.request.reg_groups)
+        elif arg_type == Environment:
+            param = Environment(self.request.environment)
         elif arg_type == Header:
             param = self.__build_header(arg, **kws)
         elif inspect.isclass(arg_type) and issubclass(arg_type, cookies.BaseCookie):
@@ -272,14 +271,14 @@ class FilterContex:
         elif arg_type in (dict, Dict):
             param = self.__build_dict(arg, **kws)
         elif type_check:
-            raise HttpError(400, f"Parameter[{arg}] with Type {arg_type} is not supported yet.")
+            raise HttpError(400, None, f"Parameter[{arg}] with Type {arg_type} is not supported yet.")
         else:
             param = val
         return param
 
     def __build_reg_group(self, val: RegGroup = RegGroup(group=0)):
         if val.group >= len(self.request.reg_groups):
-            raise HttpError(400, f"RegGroup required an element at {val.group}, but the reg length is only {len(self.request.reg_groups)}")
+            raise HttpError(400, None, f"RegGroup required an element at {val.group}, but the reg length is only {len(self.request.reg_groups)}")
         return RegGroup(group=val.group, _value=self.request.reg_groups[val.group])
 
     def __build_model_dict(self):
@@ -292,7 +291,7 @@ class FilterContex:
         return mdict
 
     def __prepare_kwargs(self):
-        kwargs = _get_kwargs_(self.__controller.func)
+        kwargs = get_function_kwargs(self.__controller.func)
         if kwargs is None:
             return None
         kwarg_vals = {}
@@ -306,12 +305,12 @@ class FilterContex:
         if name in self.request.path_values:
             return PathValue(name=name, _value=self.request.path_values[name])
         else:
-            raise HttpError(500, f"path name[{name}] not in your url mapping!")
+            raise HttpError(500, None, f"path name[{name}] not in your url mapping!")
 
     def __build_cookie(self, key, val=None):
         name = val.name if val.name is not None and val.name != "" else key
         if val._required and name not in self.request.cookies:
-            raise HttpError(400, f"Cookie[{name}] is required.")
+            raise HttpError(400, "Missing Cookie", f"Cookie[{name}] is required.")
         if name in self.request.cookies:
             morsel = self.request.cookies[name]
             cookie = Cookie()
@@ -324,13 +323,13 @@ class FilterContex:
     def __build_multipart(self, key, val=MultipartFile()):
         name = val.name if val.name is not None and val.name != "" else key
         if val._required and name not in self.request.parameter.keys():
-            raise HttpError(400, f"Parameter[{name}] is required.")
+            raise HttpError(400, "Missing Parameter", f"Parameter[{name}] is required.")
         if name in self.request.parameter.keys():
             v = self.request.parameter[key]
             if isinstance(v, MultipartFile):
                 return v
             else:
-                raise HttpError(400, f"Parameter[{name}] should be a file.")
+                raise HttpError(400, None, f"Parameter[{name}] should be a file.")
         else:
             return val
 
@@ -339,7 +338,7 @@ class FilterContex:
             try:
                 return json.loads(self.request.parameter[key])
             except:
-                raise HttpError(400, f"Parameter[{key}] should be a JSON string.")
+                raise HttpError(400, None, f"Parameter[{key}] should be a JSON string.")
         else:
             return val
 
@@ -353,19 +352,19 @@ class FilterContex:
             try:
                 return [int(p) for p in ori_list]
             except:
-                raise HttpError(400, f"One of the parameter[{key}] is not int. ")
+                raise HttpError(400, None, f"One of the parameter[{key}] is not int. ")
         elif target_type == List[float]:
             try:
                 return [float(p) for p in ori_list]
             except:
-                raise HttpError(400, f"One of the parameter[{key}] is not float. ")
+                raise HttpError(400, None, f"One of the parameter[{key}] is not float. ")
         elif target_type == List[bool]:
             return [p.lower() not in ("0", "false", "") for p in ori_list]
         elif target_type in (List[dict], List[Dict]):
             try:
                 return [json.loads(p) for p in ori_list]
             except:
-                raise HttpError(400, f"One of the parameter[{key}] is not JSON string. ")
+                raise HttpError(400, None, f"One of the parameter[{key}] is not JSON string. ")
         elif target_type == List[Parameter]:
             return [Parameter(name=key, default=p, required=False) for p in ori_list]
         else:
@@ -376,7 +375,7 @@ class FilterContex:
             try:
                 return float(self.request.parameter[key])
             except:
-                raise HttpError(400, f"Parameter[{key}] should be an float. ")
+                raise HttpError(400, None, f"Parameter[{key}] should be an float. ")
         else:
             return val
 
@@ -385,7 +384,7 @@ class FilterContex:
             try:
                 return int(self.request.parameter[key])
             except:
-                raise HttpError(400, f"Parameter[{key}] should be an int. ")
+                raise HttpError(400, None, f"Parameter[{key}] should be an int. ")
         else:
             return val
 
@@ -407,13 +406,13 @@ class FilterContex:
     def __build_json_body(self):
         if "content-type" not in self.request._headers_keys_in_lowcase.keys() or \
                 not self.request._headers_keys_in_lowcase["content-type"].lower().startswith("application/json"):
-            raise HttpError(400, 'The content type of this request must be "application/json"')
+            raise HttpError(400, None, 'The content type of this request must be "application/json"')
         return JSONBody(self.request.json)
 
     def __build_header(self, key, val=Header()):
         name = val.name if val.name is not None and val.name != "" else key
         if val._required and name not in self.request.headers:
-            raise HttpError(400, f"Header[{name}] is required.")
+            raise HttpError(400, "Missing Header", f"Header[{name}] is required.")
         if name in self.request.headers:
             v = self.request.headers[name]
             return Header(name=name, default=v, required=val._required)
@@ -423,7 +422,7 @@ class FilterContex:
     def __build_params(self, key, val=Parameters()):
         name = val.name if val.name is not None and val.name != "" else key
         if val._required and name not in self.request.parameters:
-            raise HttpError(400, f"Parameter[{name}] is required.")
+            raise HttpError(400, "Missing Parameter", f"Parameter[{name}] is required.")
         if name in self.request.parameters:
             v = self.request.parameters[name]
             return Parameters(name=name, default=v, required=val._required)
@@ -433,7 +432,7 @@ class FilterContex:
     def __build_param(self, key, val=Parameter()):
         name = val.name if val.name is not None and val.name != "" else key
         if val._required and name not in self.request.parameter:
-            raise HttpError(400, f"Parameter[{name}] is required.")
+            raise HttpError(400, "Missing Parameter", f"Parameter[{name}] is required.")
         if name in self.request.parameter:
             v = self.request.parameter[name]
             return Parameter(name=name, default=v, required=val._required)
@@ -441,59 +440,28 @@ class FilterContex:
             return val
 
 
-def _get_args_(func):
-    argspec = inspect.getfullargspec(func)
-    # ignore first argument like `self` or `clz` in object methods or class methods
-    start = 1 if inspect.ismethod(func) else 0
-    if argspec.defaults is None:
-        args = argspec.args[start:]
-    else:
-        args = argspec.args[start: len(argspec.args) - len(argspec.defaults)]
-    arg_turples = []
-    for arg in args:
-        if arg in argspec.annotations:
-            ty = argspec.annotations[arg]
-        else:
-            ty = str
-        arg_turples.append((arg, ty))
-    return arg_turples
-
-
-def _get_kwargs_(func):
-    argspec = inspect.getfullargspec(func)
-    if argspec.defaults is None:
-        return None
-
-    kwargs = OrderedDict(zip(argspec.args[-len(argspec.defaults):], argspec.defaults))
-    kwarg_turples = []
-    for k, v in kwargs.items():
-        if k in argspec.annotations:
-            k_anno = argspec.annotations[k]
-        else:
-            k_anno = str
-        kwarg_turples.append((k, v, k_anno))
-    return kwarg_turples
-
-
 class HTTPRequestHandler:
 
-    def __init__(self, base_http_quest_handler) -> None:
+    def __init__(self, base_http_quest_handler, environment = {}) -> None:
         self.base_http_quest_handler = base_http_quest_handler
         self.method = base_http_quest_handler.command
+        self.request_path = base_http_quest_handler.request_path
+        self.query_string = base_http_quest_handler.query_string
+        self.query_parameters = base_http_quest_handler.query_parameters
         self.server = base_http_quest_handler.server
         self.headers = base_http_quest_handler.headers
         self.rfile = base_http_quest_handler.rfile
         self.send_header = base_http_quest_handler.send_header
-        self.send_response = base_http_quest_handler.send_response
-        self.date_time_string = base_http_quest_handler.date_time_string
         self.end_headers = base_http_quest_handler.end_headers
+        self.send_response = base_http_quest_handler.send_response
+        self.send_error = base_http_quest_handler.send_error
         self.wfile = base_http_quest_handler.wfile
-        self.__decode_query_string = base_http_quest_handler._decode_query_string
-        self.__put_to = base_http_quest_handler._put_to
-        self.__break = base_http_quest_handler._break
-        self.query_string = base_http_quest_handler.query_string
-        self.query_parameters = base_http_quest_handler.query_parameters
-        self.request_path = base_http_quest_handler.request_path
+        self.environment = environment
+
+        self.__decode_query_string = utils.decode_query_string
+        self.__put_to = utils.put_to
+        self.__break = utils.break_into
+        self.date_time_string = utils.date_time_string
 
     def handle(self):
         mth = self.method.upper()
@@ -505,29 +473,23 @@ class HTTPRequestHandler:
 
         res = ResponseWrapper(self)
         if ctrl is None:
-            res.status_code = 404
-            res.body = {"error": "Cannot find a controller for your path"}
-            res.send_response()
+            res.send_error(404, "Controller Not Found", "Cannot find a controller for your path")
         else:
             filters = self.server.get_matched_filters(req.path)
             ctx = FilterContex(req, res, ctrl, filters)
             try:
                 ctx.do_chain()
             except HttpError as e:
-                res.status_code = e.code
-                res.body = {"error": e.message}
-                res.send_response()
+                res.send_error(e.code, e.message, e.explain)
             except Exception as e:
                 _logger.exception("error occurs! returning 500")
-                res.status_code = 500
-                res.body = {"error":  str(e)}
-                res.send_response()
+                res.send_error(500, None, str(e))
 
     def __prepare_request(self, method) -> RequestWrapper:
         path = self.request_path
         req = RequestWrapper()
+        req.environment = self.environment or {}
         req.path = "/" + path
-
         req._path = path
         headers = {}
         _headers_keys_in_lowers = {}
@@ -620,11 +582,19 @@ class HTTPRequestHandler:
         return self.__break(txt, "\r\n")
 
     def _send_response(self, response):
-        headers = response["headers"]
-        cks = response["cookies"]
+        try:
+            headers = response["headers"]
+            cks = response["cookies"]
+            raw_body = response["body"]
+            status_code = response["status_code"]
+            content_type, body = utils.decode_response_body(raw_body)
 
-        status_code, content_type, body = self.__prepare_res_body(response)
+            self._send_res(status_code, headers, content_type, cks, body)
 
+        except HttpError as e:
+            self.send_error(e.code, e.message, e.explain)
+
+    def _send_res(self, status_code: int, headers: Dict[str, str] = {}, content_type: str = "", cks: Cookies = Cookies(), body: Union[str, bytes, StaticFile] = None):
         if "Content-Type" not in headers.keys() and "content-type" not in headers.keys():
             headers["Content-Type"] = content_type
 
@@ -665,40 +635,3 @@ class HTTPRequestHandler:
                 while data:
                     self.wfile.write(data)
                     data = in_file.read(buffer_size)
-
-    def __prepare_res_body(self, response: Response):
-        status_code = response["status_code"]
-        raw_body = response["body"]
-        content_type = "text/plain; chartset=utf8"
-        if raw_body is None:
-            body = ""
-        elif isinstance(raw_body, dict):
-            content_type = "application/json; charset=utf8"
-            body = json.dumps(raw_body, ensure_ascii=False)
-        elif isinstance(raw_body, str):
-            body = raw_body.strip()
-            if body.startswith("<?xml") and body.endswith(">"):
-                content_type = "text/xml; charset=utf8"
-            elif body.lower().startswith("<!doctype html") and body.endswith(">"):
-                content_type = "text/html; charset=utf8"
-            elif body.lower().startswith("<html") and body.endswith(">"):
-                content_type = "text/html; charset=utf8"
-            else:
-                content_type = "text/plain; charset=utf8"
-        elif isinstance(raw_body, StaticFile):
-            if not os.path.isfile(raw_body.file_path):
-                _logger.error(f"Cannot find file[{raw_body.file_path}] specified in StaticFile body.")
-                status_code = 404
-                content_type = "application/json; charset=utf8"
-                body = json.dumps({
-                    "error": "Cannot find file for this url."
-                }, ensure_ascii=False)
-            else:
-                body = raw_body
-                content_type = body.content_type
-        elif isinstance(raw_body, bytes):
-            body = raw_body
-            content_type = "application/octet-stream"
-        else:
-            body = raw_body
-        return status_code, content_type, body
